/***************************************************************************
                          ibdatabase.cpp  -  description
                             -------------------
    begin                : Fri May 5 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ibdatabase.h"

IBDatabase::IBDatabase():
  Params(),
  _Connected(false),
  _DBName(),
  _Handle(NULL),
  _Dialect(1)
{
}

IBDatabase::IBDatabase(const char* DBName):
  Params(),
  _Connected(false),
  _DBName(),
  _Handle(NULL),
  _Dialect(1)
{
  setDBName( DBName );
}

IBDatabase::IBDatabase(const char* DBName, const char* User,
  const char* Password, const char* CharSet = "win1251"):
  Params(),
  _Connected(false),
  _DBName(),
  _Handle(NULL),
  _Dialect(1)
{
  setDBName( DBName );
  Params[ isc_dpb_user_name ] = User;
  Params[ isc_dpb_password  ] = Password;
  Params[ isc_dpb_lc_ctype  ] = CharSet;
}

// 
IBDatabase::~IBDatabase(){
  if( isActive() ) close();
}

// true -    ,  false
const bool IBDatabase::isActive() {
  return _Connected;
}

//    
void IBDatabase::setDBName(const char* DBName) {
  if( isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't set new DB name. Database is active");
  _DBName = DBName;
}

//     
const char* IBDatabase::getDBName() {
  return _DBName.c_str();
}

void IBDatabase::open() {
  if( isActive() ) throw IBException(__FILE__, __LINE__,
    "Database already active");
  connect();
  _Connected = true;
}

void IBDatabase::close() {
  if( !isActive() ) return;

  disconnect();
  _Connected = false;
}

isc_db_handle* IBDatabase::getDBHandle(){
  return &_Handle;
}

const unsigned short IBDatabase::getSQLDialect() {
  return _Dialect;
}

void IBDatabase::setSQLDialect(const unsigned short Dialect) {
  if(Dialect < 1 || Dialect > 3) throw IBException(__FILE__, __LINE__,
    "Can't set SQL dialect. Unsupported SQL dialect.");
  _Dialect = Dialect;
}

void IBDatabase::connect(){
  char *dpb_buffer = new char[256];
  short length;
  ISC_STATUS status_vector[20];

  length = setDPB(&dpb_buffer);
  if( _DBName.size() <= 0 ) throw IBException (__FILE__, __LINE__,
    "Database name is empty");
  _Handle = NULL;
  if( isc_attach_database( status_vector,
                           0,
                           (char*)_DBName.c_str(),
                           &_Handle,
                           length,
                           dpb_buffer ) )
    throw IBException( __FILE__, __LINE__, status_vector);
  delete[] dpb_buffer;
#ifdef IBASE_V60
  char res_info[40], *p, item;
  char db_info[] = {isc_info_db_SQL_dialect, /*isc_info_user_names,*/isc_info_end};
  if( isc_database_info(status_vector,
                        &_Handle,
                        sizeof(db_info),
                        db_info,
                        sizeof(res_info),
                        res_info ) )
    throw IBException(__FILE__, __LINE__, status_vector  );
  for( p = res_info; *p != isc_info_end;  ) {
    item = *p++;
    length = isc_vax_integer( p, 2 );
    p += 2;
    switch( item ) {
      case isc_info_db_SQL_dialect:
        _Dialect = isc_vax_integer(p, length);
      break;
    }
    p += length;
  }
#endif
}

const short IBDatabase::setDPB(char** dpb_buffer) {
  char* pBuffer;
  char* dpb;

  dpb = pBuffer = *dpb_buffer;
  *dpb++ = isc_dpb_version1;
  if( Params.find(isc_dpb_user_name) == Params.end() )
    throw IBException( __FILE__, __LINE__, "User name is empty" );
  if( Params.find( isc_dpb_password ) == Params.end() )
    throw IBException( __FILE__, __LINE__, "User password is empty" );
  for( DBParamMap::iterator prm = Params.begin(); prm != Params.end(); ++prm) {
    *dpb++ = (*prm).first;
    *dpb++ = strlen((*prm).second);
    for( const char* p = (*prm).second; *p; p++ ) *dpb++ = *p;
  }
  return (dpb - pBuffer);
}

void IBDatabase::disconnect(){
  ISC_STATUS status_vector[20];

  if( isc_detach_database( status_vector,
                           &_Handle )
    ) throw IBException( __FILE__, __LINE__, status_vector );
}