/***************************************************************************
                          ibeventcontainer.cpp  -  description
                             -------------------
    begin                : Thu May 18 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ibeventcontainer.h"

#define MAXEVENTCOUNT 15

IBEventContainer::IBEventContainer(IBDatabase* DBase):
  _EventBuf(NULL),
  _ResultBuf(NULL),
  _EventId(0),
  _BLength(0),
  _Flag(0),
  _First(true),
  _DBase(DBase),
  _Count(-1)
{
  for( short i = 0; i < MAXEVENTCOUNT; i++ ) _EventNames[i] = NULL;
}

IBEventContainer::~IBEventContainer() {
  stop();
  for( short i = 0; i < MAXEVENTCOUNT && _EventNames[i] != NULL; i++ )
    delete[] _EventNames[i];
  if(_EventBuf != NULL) delete[] _EventBuf;
  if(_ResultBuf != NULL) delete[] _ResultBuf;
}

const bool IBEventContainer::add(const char* EventName) {
   if( _Count >= MAXEVENTCOUNT || isExists(EventName) ) return false;
   _EventNames[++_Count] = new char[ strlen(EventName) + 1 ];
   strcpy(_EventNames[_Count], EventName);
   *(_EventNames[_Count] + strlen(EventName)) = '\0';
   if( _EventId ) {
      stop(); start();
   }
   return true;
}

void IBEventContainer::remove(const char* EventName) {
  short i;
  for( i = 0; i < MAXEVENTCOUNT && _EventNames[i] != NULL; i++ )
    if( strcasecmp(EventName, _EventNames[i]) == 0 ) break;
  if( i < MAXEVENTCOUNT && _EventNames[i] != NULL ) {
    delete[] _EventNames[i];
    _EventNames[i] = NULL;
    if( _EventId ) {
      stop(); start();
    }
  }
}

const bool IBEventContainer::isExists( const char* EventName ) {
  short i;
  for( i = 0; i < MAXEVENTCOUNT && _EventNames[i] != NULL; i++ )
    if( strcasecmp(EventName, _EventNames[i]) == 0 ) break;
  return (i < MAXEVENTCOUNT && _EventNames[i] != NULL);
}

isc_callback event_func(IBEventContainer* event, short length, char* updated) {
  char *r = event->_ResultBuf;
  event->_Flag++;
	while (length--) *r++ = *updated++;
	return 0;
}

void IBEventContainer::start() {
  ISC_STATUS status_vector[20];

  if( _DBase == NULL || !_DBase->isActive() ) throw IBException(__FILE__,
    __LINE__, "Can't start event dispatcher. Database not active");
  if( !_EventId )
    _BLength = isc_event_block( &_EventBuf,
                                &_ResultBuf,
                                _Count + 1,
                                _EventNames[0],
                                _EventNames[1],
                                _EventNames[2],
                                _EventNames[3],
                                _EventNames[4],
                                _EventNames[5],
                                _EventNames[6],
                                _EventNames[7],
                                _EventNames[8],
                                _EventNames[9],
                                _EventNames[10],
                                _EventNames[11],
                                _EventNames[12],
                                _EventNames[13],
                                _EventNames[14],
                                _EventNames[15] );
  if( isc_que_events(status_vector,
                     _DBase->getDBHandle(),
                     &_EventId,
                     _BLength,
                     _EventBuf,
                     (isc_callback) event_func,
                     this) )
    throw IBException(__FILE__, __LINE__, status_vector);
}

void IBEventContainer::stop() {
  ISC_STATUS status_vector[20];

  if( _EventId <= 0 ) return;
  if( isc_cancel_events(status_vector,
                        _DBase->getDBHandle(),
                        &_EventId) )
    throw IBException(__FILE__, __LINE__, status_vector);
  _EventId = 0;
}

int IBEventContainer::check() {
  unsigned long counts[20];
  int res = 0;

  if( !_Flag ) return res;
  _Flag = 0;
  isc_event_counts(counts,
                   _BLength,
                   _EventBuf,
                   _ResultBuf);
  if( _First ) _First = false;
  else for( short i = 0; i < MAXEVENTCOUNT && _EventNames[i] != NULL; i++ )
  if( counts[i] > 0 ) res = execute(_DBase, _EventNames[i], counts[i]);
  start();
  return res;
}