/***************************************************************************
                          ibquery.cpp  -  
                             -------------------
    begin                : Wed May 10 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ibquery.h"

IBQuery::IBQuery():
  Fields(this),
  Params(this),
  _SQL(NULL),
  _Cursor(NULL),
  _Active(false),
  _Prepared(false),
  _Trans(NULL),
  _LastResult(0),
  _Handle(NULL),
  _SqlType(sqlUNKNOWN),
  _Bof(true)
{}

IBQuery::IBQuery( IBTransaction* Trans ):
  Fields(this),
  Params(this),
  _SQL(NULL),
  _Cursor(NULL),
  _Active(false),
  _Prepared(false),
  _Trans(NULL),
  _LastResult(0),
  _Handle(NULL),
  _SqlType(sqlUNKNOWN),
  _Bof(true)

{
  setTransaction(Trans);
}

IBQuery::IBQuery( IBTransaction* Trans, const char* sql ):
  Fields(this),
  Params(this),
  _SQL(NULL),
  _Cursor(NULL),
  _Active(false),
  _Prepared(false),
  _Trans(NULL),
  _LastResult(0),
  _Handle(NULL),
  _SqlType(sqlUNKNOWN),
  _Bof(true)

{
  setTransaction(Trans);
  setSQL(sql);
}

IBQuery::~IBQuery(){
  close();
}

const bool IBQuery::isPrepared(){
  return _Prepared;
}

const bool IBQuery::isActive(){
  return _Active;
}

void IBQuery::prepare() {
  ISC_STATUS status_vector[20];
  int i, n = 0;
  char type_item, res_buffer[8];
  char* sql;
  short stmt_len;
  FieldNameVec FNames;
  XSQLDA* sqlda = NULL;

//  if( isPrepared() == true ) return;
  if( _Handle != NULL ) throw IBException( __FILE__, __LINE__,
    "Can't allocate statement. Query is open" );
  if ( _Trans == NULL && !_Trans->isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't allocate statement. Transaction isn't active");
  if( strlen(_SQL) <= 0 ) throw IBException(__FILE__, __LINE__,
    "Can't allocate statement. SQL is empty");
  sql = parseSQL(_SQL, &FNames);
  if( sql == NULL ) throw IBException(__FILE__, __LINE__,
    "Can't allocate statement. Wrong SQL body");

  if( isc_dsql_alloc_statement2(status_vector,
                                _Trans->getDatabase()->getDBHandle(),
                                &_Handle) )
    throw IBException(__FILE__, __LINE__, status_vector);
  if( isc_dsql_prepare(status_vector,
                       _Trans->getTRHandle(),
                       &_Handle,
                       0,
                       sql,
                       getSQLDialect(),
                       NULL) )
    throw IBException(__FILE__, __LINE__, status_vector);

  type_item = (char)isc_info_sql_stmt_type;
  if( isc_dsql_sql_info(status_vector,
                        &_Handle,
                        1,
                        &type_item,
                        sizeof(res_buffer),
                        res_buffer) )
    throw IBException(__FILE__, __LINE__, status_vector);
  if( res_buffer[0] != isc_info_sql_stmt_type ) throw IBException(__FILE__, __LINE__,
    "Can't allocate statement. Unknown statement type");
  stmt_len = isc_vax_integer(res_buffer + 1, sizeof(short) );
  _SqlType = (IBSQLType)isc_vax_integer(res_buffer+3, stmt_len);

  switch( _SqlType ) {
    case sqlGET_SEGMENT:
    case sqlPUT_SEGMENT:
    case sqlSTART_TRANSACTION:
      throw IBException(__FILE__, __LINE__,
        "Can't prepare statment. It has unsupported type.");
    break;
    case sqlCOMMIT:
    case sqlROLLBACK:
    case sqlDDL:
    case sqlINSERT:
    case sqlUPDATE:
    case sqlSELECT:
    case sqlDELETE:
    case sqlSET_GENERATOR:
    case sqlEXEC_PROCEDURE:
    case sqlSELECT_FOR_UPDATE:
      sqlda = (XSQLDA*)malloc( XSQLDA_LENGTH(1) );
      sqlda->version = SQLDA_VERSION1;
      sqlda->sqln = 1;
      if( isc_dsql_describe_bind(status_vector,
                                 &_Handle,
                                 getSQLDialect(),
                                 sqlda) )
         throw IBException(__FILE__, __LINE__, status_vector);
      if( sqlda->sqld <= 0 ) free(sqlda);
      else {
        if( sqlda->sqln < sqlda->sqld )  {
          n = sqlda->sqld;
          free(sqlda);
          sqlda = (XSQLDA*)malloc( XSQLDA_LENGTH(n) );
          sqlda->version = SQLDA_VERSION1;
          sqlda->sqln = n;
    	      if( isc_dsql_describe_bind(status_vector,
                                     &_Handle,
                                     getSQLDialect(),
                                     sqlda) )
            throw IBException(__FILE__, __LINE__, status_vector);
        }
        Params.setAsXSQLDA(sqlda);
        Params.allocateSqlData();
  	    for( i = 0; (unsigned int)i < FNames.size() && i < sqlda->sqld; i++ )
    	      Params[i].setName(FNames[i]->c_str());
    	  }
      sqlda = NULL; n = 0;
      if( _SqlType == sqlSELECT ||
          _SqlType == sqlSELECT_FOR_UPDATE ||
          _SqlType == sqlEXEC_PROCEDURE)
      {
        sqlda = (XSQLDA*)malloc( XSQLDA_LENGTH(1) );
        sqlda->version = SQLDA_VERSION1;
        sqlda->sqln = 1;
        if( isc_dsql_describe(status_vector,
                              &_Handle,
                              getSQLDialect(),
                              sqlda) )
           throw IBException(__FILE__, __LINE__, status_vector);
        if( sqlda->sqld <= 0 ) free(sqlda);
        else {
          if( sqlda->sqln < sqlda->sqld ) {
            n = sqlda->sqld;
            free(sqlda);
            sqlda = (XSQLDA*)malloc( XSQLDA_LENGTH(n) );
            sqlda->version = SQLDA_VERSION1;
            sqlda->sqln = n;
            if( isc_dsql_describe(status_vector,
                                  &_Handle,
                                  getSQLDialect(),
                                  sqlda) )
              throw IBException(__FILE__, __LINE__, status_vector);
          }
          Fields.setAsXSQLDA(sqlda);
          Fields.allocateSqlData();
        }
      }
    break;
    default: throw IBException(__FILE__, __LINE__,
      "Can't prepare statment. It has unknown type.");
  }
  if( sql != NULL ) free(sql);
  if( FNames.size() > 0 ) {
    for( FieldNameVec::iterator nm = FNames.begin(); nm != FNames.end(); ++nm )
      delete (*nm);
    FNames.clear();
  }
  _Prepared = true;
}

char* IBQuery::parseSQL(const char* sql, FieldNameVec* pNames){
  char *d, *s;
  char* ibsql;
  int difcnt = 0;
  bool inString = false;

  if( sql == NULL ) return NULL;
  pNames->clear();
  ibsql = (char*)malloc(strlen(sql) + 1);
  d = ibsql;
  for ( s = (char*)sql; s <= (sql + strlen(sql)); s++ ){
    if( *s == '\'' || *s == '\"' ) inString = !inString;
    if (*s != ':') *d++ = *s;
    else if( !inString ) {
      int plen = 0;
      char *name, *nb, *ne;

      *d++ = '?';
      nb = ++s;
      while (isalnum(*s) || *s == '_') {
        s++;
        plen++;
      }
      ne = s;
      *d++ = *s;
      if( plen <= 0 ) throw IBException(__FILE__, __LINE__,
        "Can't parse SQL. Parameter name expected but not found");
      difcnt += plen;
      name = nb;
      for( name = nb; name < nb + plen; name++ ) *name = toupper( *name );
      pNames->push_back(new string(nb, plen));
    } else *d++ = *s;
  }
  ibsql = (char*)realloc( ibsql, strlen(sql) - difcnt + 1 );
  *(ibsql + (strlen(sql) - difcnt)) = '\0';
  return ibsql;
}

void IBQuery::setTransaction(IBTransaction* TR) {
  if( isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't set transaction. Query is active");
  _Trans = TR;
}

const IBTransaction* IBQuery::getTransaction() {
  return _Trans;
}

//     
void IBQuery::close(){
  ISC_STATUS status_vector[20];

  if( !isActive() ) return;
  if( _Handle != NULL )
    if( isc_dsql_free_statement(status_vector,
                                &_Handle,
                                DSQL_drop) )
       throw IBException(__FILE__, __LINE__, status_vector);

  _Handle = NULL;
  _SqlType = sqlUNKNOWN;
  if( _Cursor != NULL ) {
    free(_Cursor);
    _Cursor = NULL;
  }
  if( _SQL != NULL ) {
    free(_SQL);
    _SQL = NULL;
  }
  Params.setCount(0);
  Fields.setCount(0);
  _LastResult = 0;
  _Prepared = false;
  _Active = false;
}

void IBQuery::setSQL(const char* sql){
  if( isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't set SQL. Query is active");
  if( _SQL != NULL ) free(_SQL);
  _SQL = (char*)malloc(strlen(sql) + 1);
  strcpy(_SQL, sql);
  *(_SQL + strlen(sql)) = '\0';
}

const char* IBQuery::getSQL() {
  return _SQL;
}

void IBQuery::execQuery() {
  ISC_STATUS status_vector[20];

  if( !isPrepared() ) prepare();
  if( _Trans == NULL || !_Trans->isActive() ) IBException(__FILE__, __LINE__,
    "Can't execute query. Transaction is not active");
  _Bof = true;
  _LastResult = 0;
  switch( _SqlType ) {
    case sqlSELECT:
      resetCursor();
      if( isc_dsql_execute2(status_vector,
	                          _Trans->getTRHandle(),
                            &_Handle,
                            getSQLDialect(),
                            (XSQLDA*)Params.getAsXSQLDA(),
         	                  NULL) )
      {
        _LastResult = -1;
        throw IBException(__FILE__, __LINE__, status_vector);
  	    }
      createCursor();
      _Active = true;
      next();
    break;
    case sqlEXEC_PROCEDURE:
      resetCursor();
      if( isc_dsql_execute2(status_vector,
	                          _Trans->getTRHandle(),
                            &_Handle,
                            getSQLDialect(),
                            Params.getAsXSQLDA(),
	                          Fields.getAsXSQLDA()) )
      {
        _LastResult = -1;
        throw IBException(__FILE__, __LINE__, status_vector);
      }
    break;
    default:
	    if( isc_dsql_execute(status_vector,
                           _Trans->getTRHandle(),
                           &_Handle,
                           getSQLDialect(),
                           Params.getAsXSQLDA()) )
      {
        _LastResult = -1;
        throw IBException(__FILE__, __LINE__, status_vector);
      }
    break;
  }
  _Active = true;
}

void IBQuery::next() {
  if( _LastResult != 0 || _Handle == NULL ) throw IBException(__FILE__, __LINE__,
    "Can't go to next. Query is not intialized");
  if( !isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't go to next. Query is not executed");
  _LastResult = isc_dsql_fetch(_FetchStatus,
                               &_Handle,
                               getSQLDialect(),
                               Fields.getAsXSQLDA());
  if (_FetchStatus[0] == 1 && _FetchStatus[1]){
    _LastResult  = -1;
    throw IBException(__FILE__, __LINE__, _FetchStatus);
  }
  if( _Bof && !isEOF()  ) _Bof = false;
}

const bool IBQuery::isEOF() {
  if( _LastResult == 0 ) return false;
  if( _LastResult != 100L ) throw IBException(__FILE__, __LINE__, _FetchStatus);
  return _LastResult == 100L;
}

void IBQuery::resetCursor() {
  ISC_STATUS status_vector[20];

  if( _Cursor == NULL ) return;
  if( isc_dsql_free_statement(status_vector,
                              &_Handle,
                              DSQL_close) )
  {
    _LastResult = -1;
      throw  IBException(__FILE__, __LINE__, status_vector);
  }
  if( _Cursor != NULL ) {
    free(_Cursor);
    _Cursor = NULL;
  }
//  createCursor();
}

void IBQuery::createCursor() {
  ISC_STATUS status_vector[20];

  if( _Cursor != NULL ) free(_Cursor);
  _Cursor = (char*)malloc(32);
  sprintf(_Cursor, "Stmt_Hamdle=%x", (int)_Handle);
  if( isc_dsql_set_cursor_name(status_vector,
                               &_Handle,
                               _Cursor,
                               0) )
    throw IBException(__FILE__, __LINE__, status_vector);
}

const bool IBQuery::isEmpty() {
  return _Bof && isEOF();
}

const unsigned short IBQuery::getSQLDialect() {
  return ((IBDatabase*)_Trans->getDatabase())->getSQLDialect();
}