/***************************************************************************
                          ibquery.h  -  SQL 
                             -------------------
    begin                : Wed May 10 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef IBQUERY_H
#define IBQUERY_H

extern "C" {
#include <ibase.h>
#include <stdlib.h>
}

#include "ibxsqlda.h"
#include "ibtransaction.h"
#include "ibdatabase.h"
#include "ibexception.h"
#include "ibasepp_common.h"

using namespace std;

/***************************************************************************
   SQL -    ibasepp
 ***************************************************************************/

class IBQuery {
public:
	IBQuery();
  IBQuery( IBTransaction* Trans );
  IBQuery( IBTransaction* Trans, const char* sql );
	~IBQuery();
public: // fields
  enum IBSQLType {
    sqlUNKNOWN = 0, sqlSELECT, sqlINSERT, 	sqlUPDATE, sqlDELETE, sqlDDL,
    sqlGET_SEGMENT, sqlPUT_SEGMENT, sqlEXEC_PROCEDURE, sqlSTART_TRANSACTION,
    sqlCOMMIT, sqlROLLBACK, sqlSELECT_FOR_UPDATE, sqlSET_GENERATOR
  };
	IBXSQLDA Fields;
	IBXSQLDA Params;
public: // methods	
	void prepare();
	const bool isActive();
	const bool isPrepared();
	void setTransaction(IBTransaction* Trans);
	const IBTransaction* getTransaction();
	void close();
  void setSQL(const char* sql);
  const char* getSQL();
  void execQuery();
  void next();
  const bool isEOF();
  const bool isEmpty();
private:
  typedef vector<string*> FieldNameVec;
private: // fields
  char* _SQL;
  char* _Cursor;
  bool _Active;
  bool _Prepared;
  IBTransaction* _Trans;
  ISC_STATUS _FetchStatus[20];
  ISC_STATUS _LastResult;
  isc_stmt_handle _Handle;
  IBSQLType _SqlType;
  bool _Bof;
private: //methods
  char* parseSQL(const char* sql, FieldNameVec* names);
  void createCursor();
  void resetCursor();
  const unsigned short getSQLDialect();
};

#endif
