/***************************************************************************
                          ibtransaction.cpp  -  description
                             -------------------
    begin                : Fri May 5 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ibtransaction.h"

const short TPBParamNameCount = 20;

const char* TPBParamNames[] = {
  "consistency"     ,
  "concurrency"     ,
  "shared"          ,
  "protected"       ,
  "exclusive"       ,
  "wait"            ,
  "nowait"          ,
  "read"            ,
  "write"           ,
  "lock_read"       ,
  "lock_write"      ,
  "verb_time"       ,
  "commit_time"     ,
  "ignore_limbo"    ,
  "read_committed"  ,
  "autocommit"      ,
  "rec_version"     ,
  "no_rec_version"  ,
  "restart_requests",
  "no_auto_undo"
};

IBTransaction::IBTransaction():
  _Handle(NULL),
  _DBase(NULL),
  _Active(false),
  _Params()
{
  setDefaultTPB();
}

IBTransaction::IBTransaction(IBDatabase* DBase):
  _Handle(NULL),
  _DBase(NULL),
  _Active(false),
  _Params()
{
  setDatabase(DBase);
  setDefaultTPB();
}

// 
IBTransaction::~IBTransaction(){
  commit();
}

void IBTransaction::clearParams() {
  if( isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't clear TPB parameters. Transaction is active");
  _Params.clear();
}

void IBTransaction::setParam(const char* tpb_param) {
  if( isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't set TPB parameter. Transaction is active");
  _Params.push_back(tpb_param);
}

// open transaction
void IBTransaction::startTransaction(){
  ISC_STATUS status_vector[20];
  unsigned short len;
  char *buffer = new char[128];

  if( isActive() ) throw IBException( __FILE__, __LINE__,
    "Can't start transaction. Transaction already active" );
  if( _DBase == NULL ) throw IBException( __FILE__, __LINE__,
    "Can't start transaction. Database not assigned" );
  if( !_DBase->isActive() ) throw IBException( __FILE__, __LINE__,
    "Can't start transaction. Database not active" );
  _Handle = NULL;
  len = setTPB(&buffer);
  if( isc_start_transaction(status_vector,
                            &_Handle,
                            1,
                            _DBase->getDBHandle(),
                            (unsigned short)_Params.size(),
                            buffer) )
    throw IBException( __FILE__, __LINE__, status_vector );
  delete[] buffer;
  _Active = true;
}

//    
void IBTransaction::commit(){
  ISC_STATUS status_vector[20];

  if( !isActive() ) return;
  if( isc_commit_transaction(status_vector,
                             &_Handle) )
    throw IBException( __FILE__, __LINE__, status_vector );
  _Active = false;
}

//   
void IBTransaction::rollback(){
  ISC_STATUS status_vector[20];

  if( !isActive() ) return;
  if( isc_rollback_transaction(status_vector,
                               &_Handle) )
    throw IBException( __FILE__, __LINE__, status_vector );
  _Active = false;
}

//    
void IBTransaction::commitRetaining(){
  ISC_STATUS status_vector[20];

  if( !isActive() ) return;
  if( isc_commit_retaining(status_vector,
                           &_Handle) )
    throw IBException( __FILE__, __LINE__, status_vector );
}

//    
void IBTransaction::rollbackRetaining(){
  ISC_STATUS status_vector[20];

  if( !isActive() ) return;
  if( isc_rollback_retaining(status_vector,
                           &_Handle) )
    throw IBException( __FILE__, __LINE__, status_vector );
}

//   
const bool IBTransaction::isActive(){
  return _Active;
}

//    
isc_tr_handle* IBTransaction::getTRHandle(){
  return &_Handle;
}

//    
IBDatabase* IBTransaction::getDatabase(){
  return _DBase;
}

void IBTransaction::setDatabase( IBDatabase* DBase ) {
  if( isActive() ) throw IBException(__FILE__, __LINE__,
    "Can't set database. Transaction is active");
  _DBase = DBase;
}

const unsigned short IBTransaction::setTPB(char** tpb_buffer) {
  if( _Params.size() <= 0 ) throw IBException(__FILE__, __LINE__,
    "Cant't start transaction. TPB buffer is empty.");
  char* prm_name, *prm_value;
  const char* prm_str;
  int isc_param = 0;
  char *tpb;

  tpb = *tpb_buffer;
  *tpb++ = isc_tpb_version3;
  for( TRParamVec::iterator prm = _Params.begin(); prm != _Params.end(); ++prm ) {
    prm_str = *prm;
    if( (prm_value = strstr(prm_str, "=")) != NULL ) {
      prm_name = (char*) malloc( prm_value - prm_str + 1 );
      strncpy(prm_name, prm_str, prm_value - prm_str);
      *(prm_name + (prm_value - prm_str)) = 0;
      prm_value++;
    } else {
      prm_name = (char*) malloc( strlen(prm_str) + 1 );
      strcpy(prm_name, prm_str);
      *(prm_name + strlen(prm_str)) = 0;
    }
    for( isc_param = 0; isc_param < TPBParamNameCount; isc_param++ )
      if( strcasecmp( TPBParamNames[isc_param], prm_name ) == 0 ) break;
    if( isc_param < TPBParamNameCount  ) {
      switch( isc_param + 1 ) {
        case isc_tpb_consistency:
        case isc_tpb_exclusive:
        case isc_tpb_concurrency:
        case isc_tpb_shared:
        case isc_tpb_wait:
        case isc_tpb_nowait:
        case isc_tpb_read:
        case isc_tpb_write:
        case isc_tpb_ignore_limbo:
        case isc_tpb_read_committed:
        case isc_tpb_rec_version:
        case isc_tpb_no_rec_version:
          *tpb++ = isc_param + 1;
        break;
        case isc_tpb_lock_read:
        case isc_tpb_lock_write:
          *tpb++ = isc_param + 1;
          *tpb++ = strlen(prm_value);
          strcpy(tpb, prm_value);
          tpb += strlen(prm_value);
        break;
        default: throw IBException( __FILE__, __LINE__, "Invalid TPB parameter" );
      }
    } else throw IBException( __FILE__, __LINE__, "TPB parameter not found" );
    free(prm_name);
  }
  *tpb++ = 0;
  return tpb - *tpb_buffer;
}

void IBTransaction::setDefaultTPB() {
  clearParams();
  setParam("concurrency");
  setParam("no_rec_version");
  setParam("nowait");
}
