/***************************************************************************
                          ibxsqlda.cpp  -  
                             -------------------
    begin                : Wed May 10 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ibxsqlda.h"

//#include "align.h"

IBXSQLDA::IBXSQLDA():
  _Sqlda(NULL),
  _Vars(),
  _Query(NULL)
{}

IBXSQLDA::IBXSQLDA(IBQuery* query):
  _Sqlda(NULL),
  _Vars(),
  _Query(NULL)
{
  setQuery( query );
}

// 
IBXSQLDA::~IBXSQLDA() {
  setCount(0);
}

void IBXSQLDA::setQuery( IBQuery* query ) {
  if( query == NULL ) return;
  _Query = query;
}

const IBQuery* IBXSQLDA::getQuery() {
  return _Query;
}

void IBXSQLDA::setCount( const unsigned int count ) {
  if( _Sqlda != NULL ) {
    for( FieldVec::iterator fi = _Vars.begin(); fi != _Vars.end(); ++fi )
      delete *fi;
    _Vars.clear();
    free(_Sqlda);
    _Sqlda = NULL;
  }
  if( count <= 0 ) return;
#ifdef IBASE_V60
  _Sqlda = (XSQLDA*)malloc( XSQLDA_LENGTH(count) );
#else
  _Sqlda = (XSQLDA*)malloc( SQLDA_LENGTH(count) );
#endif
  _Sqlda->version = SQLDA_VERSION1;
  _Sqlda->sqln = count;
}

const int IBXSQLDA::getCount() {
  if( _Sqlda == NULL ) return 0;
  return _Sqlda->sqld;
}

void IBXSQLDA::allocateSqlData() {
  int i;
  XSQLVAR* var;

  if( _Sqlda == NULL ) return;
  for( i = 0, var = _Sqlda->sqlvar; i < _Sqlda->sqld; i++, var++ ) {
    var->sqldata = NULL;
    var->sqlind = NULL;
    if( var->aliasname_length == 0 ) {
      memset(var->aliasname, 0, 32);
      snprintf(var->aliasname, 31, "F_%d", i + 1);
      var->aliasname_length = strlen(var->aliasname);
    }
    _Vars.push_back( createIBXSQLVAR( i ) );
  }
}

IBXSQLVAR* IBXSQLDA::createIBXSQLVAR( const unsigned int FieldIndex ) {
  IBXSQLVAR* field;

  if( FieldIndex < 0 ) return NULL;

  switch( _Sqlda->sqlvar[FieldIndex].sqltype & ~1 ) {
    case SQL_VARYING:
    case SQL_TEXT:
      field = new IBStringField();
    break;
    case SQL_SHORT:
    case SQL_LONG:
      field = new IBIntField();
    break;
    case SQL_FLOAT:
    case SQL_DOUBLE:
      field = new IBFloatField();
    break;
#ifndef IBASE_V60
    case SQL_DATE:
#else
    case SQL_TYPE_DATE:
    case SQL_TYPE_TIME:
    case SQL_TIMESTAMP:
#endif
      field = new IBDateTimeField();
    break;
    case SQL_BLOB:
      field = new IBBlobField();
    break;
    default: throw IBException(__FILE__, __LINE__,
      "Can't create IBField. Unsupported field type");
  }
  field->setIBXSQLDA( this );
  field->setFieldIndex( FieldIndex );
  return (IBXSQLVAR*)field;
}

IBXSQLVAR& IBXSQLDA::getByIndex(const int Index) {
  if( Index < 0 || (unsigned int)Index >= _Vars.size() )
    throw IBException(__FILE__, __LINE__,
      "Invalid IBXSQLVAR reference. Index is out of bounds");
  PIBXSQLVAR varObj = _Vars[Index];
  return *varObj;
}

IBXSQLVAR& IBXSQLDA::getByName(const char* FieldName) {
  if( FieldName == NULL ) throw IBException(__FILE__, __LINE__,
    "Invalid IBSQLVAR reference. Bad field name");
  FieldVec::iterator fi = find_if(_Vars.begin(), _Vars.end(), IBCheckField(FieldName));
  if( fi == _Vars.end() ) throw IBException(__FILE__, __LINE__,
    "Invalid IBSQLVAR reference. Field not found");
  return *(*fi);
}

IBXSQLVAR& IBXSQLDA::operator[](const int Index) {
  return getByIndex(Index);
}

IBXSQLVAR& IBXSQLDA::operator[](const char* FieldName) {
  return getByName(FieldName);
}

XSQLDA* IBXSQLDA::getAsXSQLDA() {
  return _Sqlda;
}

void IBXSQLDA::setAsXSQLDA(XSQLDA* sqlda) {
  if( sqlda == NULL ) return;
  if(_Sqlda != NULL) setCount(0);
  _Sqlda = sqlda;
}
