/***************************************************************************
                          ibxsqlvar.cpp  -  description
                             -------------------
    begin                : Sat May 6 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ibxsqlvar.h"
#include "ibxsqlda.h"
#include "ibquery.h"
#include "tables.h"

#define NULL_STR "<Null>"
#define EMPTY_STR "<Empty>"

// IBXSQLVAR

IBXSQLVAR::IBXSQLVAR():
  _Owner(NULL),
  _FieldIndex(-1),
  _StrBuffer( NULL )
{}

IBXSQLVAR::IBXSQLVAR(IBXSQLDA* sqlda, const int index):
  _Owner(NULL),
  _FieldIndex(index),
  _StrBuffer( NULL )

{
  setIBXSQLDA(sqlda);
  setFieldIndex(index);
}

IBXSQLVAR::~IBXSQLVAR(){
  clearData();
  clearStrBuffer();
}

void IBXSQLVAR::setName(const char* newname) {
  memset( _var().aliasname, 0, 32 );
  if( strlen(newname) > 31 )
    strncpy( _var().aliasname, newname, 31 );
  else
    strcpy( _var().aliasname, newname );
  _var().aliasname_length = strlen(_var().aliasname);
}

const char* IBXSQLVAR::getName() {
  return _var().aliasname;
}

void IBXSQLVAR::setIBXSQLDA( IBXSQLDA* sqlda ) {
  _Owner = sqlda;
}

void IBXSQLVAR::setFieldIndex( const int index ) {
  _FieldIndex = index;
  clearData();
  allocateData();
}

void IBXSQLVAR::allocateData() {
  int length = _var().sqllen;
  if( getType() == SQL_TEXT ) length += 1;
  else if( getType() ==  SQL_VARYING )
    length += sizeof(short) + 1;
  _var().sqldata = new char[ length ];
  if( isNullable() ) _var().sqlind = new short();
}

void IBXSQLVAR::clearData() {
  if( _FieldIndex >= 0 ) {
    if( _var().sqldata != NULL) delete[] _var().sqldata;
    if( isNullable() && _var().sqlind != NULL )
      delete _var().sqlind;
  }
}

const int IBXSQLVAR::getType() {
  return _var().sqltype & ~1;
}

const int IBXSQLVAR::getLength() {
  if( getType() == SQL_VARYING )
    return isc_vax_integer( _var().sqldata, sizeof(short));
  return _var().sqllen;
}

const XSQLVAR* IBXSQLVAR::getAsXSQLVAR() {
  return &_var();
}

const bool IBXSQLVAR::isNull() {
  return isNullable() && *_var().sqlind < 0;
}

const bool IBXSQLVAR::isNullable() {
  return (_var().sqltype & 1) > 0;
}

//  
IBXSQLVAR::operator const char*() {
  clearStrBuffer();
  _StrBuffer = (char*)getAsString();
  return _StrBuffer;
}

IBXSQLVAR::operator short() {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
}

IBXSQLVAR::operator int() {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
}

IBXSQLVAR::operator long() {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
}

IBXSQLVAR::operator float() {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
}

IBXSQLVAR::operator double() {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
}

IBXSQLVAR::operator struct tm*() {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
}

//  

IBXSQLVAR& IBXSQLVAR::operator = (const char* value) {
  if( value == NULL ) setNull();
  else setAsString(value);
  return *this;
}

IBXSQLVAR& IBXSQLVAR::operator = (const short value) {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
  return *this;
}

IBXSQLVAR& IBXSQLVAR::operator = (const int value) {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
  return *this;
}

IBXSQLVAR& IBXSQLVAR::operator = (const long value) {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
  return *this;
}

IBXSQLVAR& IBXSQLVAR::operator = (const float value) {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
  return *this;
}

IBXSQLVAR& IBXSQLVAR::operator = (const double value) {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
  return *this;
}

IBXSQLVAR& IBXSQLVAR::operator = (const struct tm* value) {
  throw IBException(__FILE__, __LINE__, "Invalid IBXSQLVAR type conversion");
  return *this;
}

IBXSQLVAR& IBXSQLVAR::operator = (IBXSQLVAR& var) {
  if( getType() != var.getType() ) throw IBException(__FILE__, __LINE__,
    "XSQLVAR unsaccessfull assignment. Types isn't equal");
  memcpy( _var().sqldata, var._var().sqldata, _var().sqllen );
  if( isNullable() )
    if( var.isNull() ) setNull();
    else *(short*)(_var().sqlind) = 0;
  return *this;
}

void IBXSQLVAR::checkFieldType(const bool Good) {
  if( !Good ) throw IBException(__FILE__, __LINE__,
    "Invalid XSLVAR type for this operation.");
}

XSQLVAR& IBXSQLVAR::_var() {
  if( _Owner == NULL ) throw IBException(__FILE__, __LINE__,
    "Bad XSQLVAR reference. Invalid IBXSQLDA.");
  XSQLDA* sqlda =  _Owner->getAsXSQLDA();
  if( _FieldIndex < 0  || _FieldIndex >= sqlda->sqld )
    throw  IBException(__FILE__, __LINE__,
      "Bad XSQLVAR reference. _FieldIndex out of bounds.");
  return sqlda->sqlvar[_FieldIndex];
}

void IBXSQLVAR::setNull() {
  if( !isNullable() ) throw IBException( __FILE__, __LINE__,
    "XSQLVAR can't be null" );
  *(short*)(_var().sqlind)= -1;
}

void IBXSQLVAR::clearStrBuffer() {
  if( _StrBuffer == NULL ) return;
  free( _StrBuffer );
  _StrBuffer = NULL;
}

// IBXSQLVAR friends

ostream& operator<<(ostream& out, IBXSQLVAR& var) {
  out << (const char*)var;
  return out;
}

// common used

char* formatStr(const char* format, ...) {
  if( format == NULL  ) return NULL;
  int n, size = 50;
  char* p;
  va_list ap;

  p = (char*)malloc(size);
  while( 1 ) {
    memset(p, 0, size);
    va_start(ap, format);
    n = vsnprintf(p, size, format, ap);
    va_end(ap);
    if( n > -1 && n < size ) break;
    if( n > -1 ) size += 1;
    else size *= 2;
    p = (char*)realloc(p, size);
  }
  return p;
}

// IBIntField

IBIntField::operator short() {
  return getAsInt();
}

IBXSQLVAR& IBIntField::operator =(const short value) {
  setAsInt(value);
  return *this;
}

IBIntField::operator int() {
  return getAsInt();
}

IBXSQLVAR& IBIntField::operator =(const int value) {
  setAsInt(value);
  return *this;
}

IBIntField::operator long() {
  return getAsInt();
}

IBXSQLVAR& IBIntField::operator =(const long value) {
  setAsInt(value);
  return *this;
}

long IBIntField::getAsInt() {
  checkFieldType( getType() == SQL_LONG  || getType() == SQL_SHORT );
  if( isNull() ) return 0;
  if( getType() == SQL_LONG ) return  *(long*)(_var().sqldata);
  else return *(short*)(_var().sqldata);
}


void IBIntField::setAsInt(const long value) {
  checkFieldType( getType() == SQL_LONG  || getType() == SQL_SHORT );
  if( isNullable() && _var().sqlind != NULL )
    *(short*)(_var().sqlind) = 0;
  if( getType() == SQL_LONG ) *(long*)(_var().sqldata) = value;
  else  *(short*)(_var().sqldata) = value;
}

char*IBIntField::getAsString() {
  if( isNullable() && isNull() ) return formatStr("%s", NULL_STR);
  else return formatStr("%d", getAsInt());
}

void IBIntField::setAsString(const char* value) {
  if( value == NULL ) setNull();
  else setAsInt( atol(value) );
}

// IBFloatField

IBFloatField::operator float() {
  return getAsFloat();
}

IBXSQLVAR& IBFloatField::operator =(const float value) {
  setAsFloat(value);
  return *this;
}

IBFloatField::operator double() {
  return getAsFloat();
}

IBXSQLVAR& IBFloatField::operator =(const double value) {
  setAsFloat(value);
  return *this;
}

double IBFloatField::getAsFloat() {
  checkFieldType( getType() == SQL_FLOAT || getType() == SQL_DOUBLE );
  if( isNull() ) return 0;
  if( getType() == SQL_FLOAT ) return  *(float*)_var().sqldata;
  else return  *(double*)_var().sqldata;
}

void IBFloatField::setAsFloat(const double value) {
  checkFieldType( getType() == SQL_FLOAT || getType() == SQL_DOUBLE );
  if( isNullable() && _var().sqlind != NULL )
    *(short*)(_var().sqlind) = 0;
  if( getType() == SQL_FLOAT ) *(float*)(_var().sqldata) = value;
  else  *(double*)(_var().sqldata) = value;
}

char*IBFloatField::getAsString() {
  if( isNullable() && isNull() ) return formatStr("%s", NULL_STR);
  else return formatStr("%f", getAsFloat());
}

void IBFloatField::setAsString(const char* value) {
  if( value == NULL ) setNull();
  else setAsFloat( atof(value) );
}

// IBStringField

char*IBStringField::getAsString() {
  char* str, *result;

  checkFieldType( getType() == SQL_TEXT || getType() == SQL_VARYING );
  if( isNull() ) str = NULL_STR;
  else if( getType() == SQL_VARYING ) str = _var().sqldata + sizeof(short);
  else str = _var().sqldata;
  result = (char*)malloc( sizeof(char) *( getLength() + 1 ) );
  strncpy( result, str, getLength() );
  *(result + getLength()) = '\0';
  if( hasWin1251() ) convertWin2Koi( &result, getLength() );
  return result;
}

void IBStringField::setAsString(const char* value) {
  checkFieldType( getType() == SQL_TEXT || getType() == SQL_VARYING );
  if( isNullable() )
    if( value == NULL ) {
      setNull();
      return;
    } else if( _var().sqlind != NULL ) *(short*)(_var().sqlind) = 0;
  if( getType() == SQL_VARYING ) _var().sqltype = SQL_TEXT;
  _var().sqllen = strlen(value);
  _var().sqldata = (char*)realloc( _var().sqldata,
                                   sizeof(char) * (_var().sqllen + 1) );
  memset( _var().sqldata, 0, _var().sqllen + 1 );
  strcpy(_var().sqldata, value);
  if( hasWin1251() ) convertKoi2Win(&_var().sqldata, _var().sqllen);
}

const bool IBStringField::hasWin1251() {
  return true;
}

/************************************************************************
     Windows  KOI8  
 ************************************************************************/
void IBStringField::convertWin2Koi(char** WinStr, int len){
  char *p = *WinStr;

  while( p <= (*WinStr + len) )
    *p++ = win2koi[(unsigned char)*p];
}

void IBStringField::convertKoi2Win(char** KoiStr, int len){
  char *p = *KoiStr;

  while( p <= (*KoiStr + len) )
    *p++ = koi2win[(unsigned char)*p];
}

// IBDateTimeField

IBDateTimeField::operator struct tm*() {
  return getAsDateTime();
}

IBXSQLVAR& IBDateTimeField::operator =(const struct tm* value) {
  setAsDateTime(value);
  return *this;
}

struct tm* IBDateTimeField::getAsDateTime() {
  struct tm* dt = &_Buffer;

#ifndef IBASE_V60
  checkFieldType( getType() == SQL_DATE );
#else
  checkFieldType( getType() == SQL_TYPE_DATE ||
                  getType() == SQL_TYPE_TIME ||
                  getType() == SQL_TIMESTAMP );
#endif
  if( isNull() ) return NULL;
  dt->tm_sec = dt->tm_min = dt->tm_hour = dt->tm_mday = dt->tm_mon =
    dt->tm_year = 0;
#ifndef IBASE_V60
    ISC_QUAD  date;
    memcpy(&date, _var().sqldata, _var().sqllen);
    isc_decode_date(&date, dt);
#else
  switch( getType() ) {
    case SQL_TYPE_DATE: {
      ISC_DATE ibdate;
      memcpy( &ibdate, _var().sqldata, _var().sqllen );
      isc_decode_sql_date( &ibdate, dt );
      break;
    }
    case SQL_TYPE_TIME: {
      ISC_TIME ibtime;
      memcpy( &ibtime, _var().sqldata, _var().sqllen );
      isc_decode_sql_time( &ibtime, dt );
      break;
    }
    case SQL_TIMESTAMP: {
      ISC_TIMESTAMP ibdatetime;
      memcpy( &ibdatetime, _var().sqldata, _var().sqllen );
      isc_decode_timestamp( &ibdatetime, dt );
      break;
    }
  }
#endif
  return  dt;
}

void IBDateTimeField::setAsDateTime(const struct tm* value) {
  if( isNullable() )
    if( value == NULL ) {
      setNull();
      return;
    } else if( _var().sqlind != NULL ) *(short*)(_var().sqlind) = 0;
#ifndef IBASE_V60
  checkFieldType( getType() == SQL_DATE );
  ISC_QUAD  date;
  isc_encode_date((struct tm*)value, &date);
  memcpy(_var().sqldata, &date, _var().sqllen);
#else
  checkFieldType( getType() == SQL_TYPE_DATE ||
                  getType() == SQL_TYPE_TIME ||
                  getType() == SQL_TIMESTAMP );
  switch( getType() ) {
    case SQL_TYPE_DATE: {
      ISC_DATE ibdate;
      isc_encode_sql_date( (struct tm*)value, &ibdate );
      memcpy( _var().sqldata, &ibdate, _var().sqllen );
      break;
    }
    case SQL_TYPE_TIME: {
      ISC_TIME ibtime;
      isc_encode_sql_time( (struct tm*)value, &ibtime );
      memcpy( _var().sqldata, &ibtime, _var().sqllen );
      break;
    }
    case SQL_TIMESTAMP: {
      ISC_TIMESTAMP ibdatetime;
      isc_encode_timestamp( (struct tm*)value, &ibdatetime );
      memcpy( _var().sqldata, &ibdatetime, _var().sqllen );
      break;
    }
  }
#endif
}

char*IBDateTimeField::getAsString() {
  char* p, *e;
  if( isNullable() && isNull() ) return formatStr("%s", NULL_STR);
  p = (char*)formatStr( "%s", asctime( getAsDateTime() ) );
  e = p + strlen(p) - 1;
  while( *e == '\n' ) e--; e++;
  p = (char*) realloc( p, e - p );
  *(p + (e-p)) = '\0';
  return p;
}

void IBDateTimeField::setAsString(const char* value) {
  struct tm date;

  if( value == NULL ) {
    setNull();
    return;
  }
  date.tm_hour = 0;
  date.tm_min  = 0;
  date.tm_sec  = 0;
  if( strchr(value, '/') ) {
// 01/01/2000 00:00:00
    if( strchr(value, ':') ) strptime(value, "%d/%m/%Y %T", &date);
    else strptime(value, "%d/%m/%Y", &date);
  } else if( strchr(value, '.') ) {
// 01.01.2000 00:00:00
    if( strchr(value, ':') ) strptime(value, "%d.%m.%Y %T", &date);
    else strptime(value, "%d.%m.%Y", &date);
  } else if( strchr(value, '-') ) {
// 01-JAN-2000 00:00:00
    if( strchr(value, ':') ) strptime(value, "%d-%b-%Y %T", &date);
    else strptime(value, "%d-%b-%Y", &date);
  } else throw IBException(__FILE__, __LINE__,
    "invalid IBDateTimeField assignment. Unknown date format");
  setAsDateTime( &date );
}

// IBBlobField

IBBlobField::IBBlobField():
  _Handle(NULL),
  _Length(0),
  _SegmentCount(0),
  _SegmentSize(0)
{}


const int IBBlobField::getSubType() {
  return _var().sqlsubtype;
}

const ISC_QUAD* IBBlobField::getBlobId() {
  checkFieldType( getType() == SQL_BLOB );
  if( isNull() ) return NULL;
  return  (ISC_QUAD *)_var().sqldata;
}

const int IBBlobField::getLength() {
  checkBlob();
  return _Length;
}

char*IBBlobField::getAsString() {
  char *seg_buf = NULL, *blob_buf = NULL;
  int offset = 0;
  unsigned short seg_len = 0;
  ISC_STATUS result, status_vector[20];

  if( isNull() ) {
    blob_buf = (char*) malloc( sizeof(char) * (strlen(NULL_STR) + 1) );
    strcpy(blob_buf, NULL_STR);
    *(blob_buf + strlen(NULL_STR)) = '\0';
    return blob_buf;
  }
  open();
  if( _Length <= 0 )  {
    blob_buf = (char*)malloc( sizeof(char*) * (strlen(EMPTY_STR) + 1) );
    strcpy(blob_buf, EMPTY_STR);
    *(blob_buf + strlen(EMPTY_STR)) = '\0';
    return blob_buf;
  }
  seg_buf = (char*) malloc(sizeof(char) * _SegmentSize);

  result = isc_get_segment(status_vector,
                           &_Handle,
                           &seg_len,
                           _SegmentSize,
                           seg_buf);
  while( result == 0 || status_vector[1] == isc_segment ) {
    if( blob_buf == NULL ) blob_buf = (char*) malloc( seg_len );
    else blob_buf = (char*) realloc( blob_buf, offset + seg_len );
    strncpy( blob_buf + offset, seg_buf, seg_len );
    offset += seg_len;
    result = isc_get_segment(status_vector,
                             &_Handle,
                             &seg_len,
                             _SegmentSize,
                             seg_buf);
    memset(seg_buf, 0, _SegmentSize);
  }
  if( blob_buf != NULL ) {
    blob_buf = (char*) realloc( blob_buf, offset + 1 );
    *(blob_buf + offset) = '\0';
  }
  close();
  delete[] seg_buf;
  if( hasWin1251() ) convertWin2Koi( &blob_buf, strlen(blob_buf) );
  return blob_buf;
}

void IBBlobField::setAsString( const char* value ) {
  ISC_STATUS status_vector[20];
  char* buf = NULL;

  if( isNullable() )
    if( value == NULL ) {
      setNull();
      return;
    } else if( _var().sqlind != NULL ) *(short*)(_var().sqlind) = 0;
  buf = new char[strlen(value) + 1];
  strcpy( buf, value );
  *(buf + strlen(value)) = '\0';
  if( hasWin1251() ) convertKoi2Win( &buf, strlen(buf) );
  create();
  if( isc_put_segment(status_vector,
                      &_Handle,
                      strlen(value),
                      (char*)value) ) {
    close();
    throw IBException(__FILE__, __LINE__, status_vector);
  }
  close();
  delete[] buf;
}

void IBBlobField::checkBlob() {
  if( getBlobId() == NULL ) throw IBException(__FILE__, __LINE__,
    "Can't do this operation with BLOB. Field value is Null" );
  if( _Handle == NULL ) throw IBException( __FILE__, __LINE__,
    "Can't do this operation with BLOB. BLOB isn't open" );
}

void IBBlobField::open() {
  ISC_STATUS status_vector[20];
  char blob_items[] = {isc_info_blob_max_segment,
                       isc_info_blob_num_segments,
                       isc_info_blob_total_length};
  char res_buffer[60], *p, item;
  int len;

  if( getBlobId() == NULL ) throw IBException(__FILE__, __LINE__,
    "Can't do this operation with BLOB. Field value is Null" );
  isc_db_handle db = ((IBDatabase*)((IBTransaction*)((IBQuery*)_Owner->getQuery())->getTransaction())->getDatabase())->getDBHandle();
  isc_tr_handle tr = ((IBTransaction*)((IBQuery*)_Owner->getQuery())->getTransaction())->getTRHandle();
  _Handle = NULL;
  if( db == NULL || tr == NULL ) throw IBException( __FILE__, __LINE__,
    "Can't allocate BLOB. Invalid database or transaction." );
  if( isc_open_blob2(status_vector,
                     (isc_db_handle*)db,
                     (isc_tr_handle*)tr,
                     &_Handle,
                     (ISC_QUAD*)getBlobId(),
                     0,
                     NULL) )
    throw IBException( __FILE__, __LINE__, status_vector );
  if( isc_blob_info(status_vector,
                    &_Handle,
                    sizeof(blob_items),
                    blob_items,
                    sizeof(res_buffer),
                    res_buffer) ) {
     close();
     throw IBException(__FILE__, __LINE__, status_vector);
  }
  for( p = res_buffer; *p != isc_info_end;  ) {
    item = *p++;
    len = (short) isc_vax_integer( p, sizeof(short) );
    p += sizeof(short);
    switch( item ) {
      case isc_info_blob_max_segment:
        _SegmentSize = isc_vax_integer(p, len);
      break;
      case isc_info_blob_num_segments:
        _SegmentCount = isc_vax_integer(p, len);
      break;
      case isc_info_blob_total_length:
        _Length = isc_vax_integer(p, len);
      break;
      default: break;
    }
    p += len;
  }
}

void IBBlobField::close(){
  ISC_STATUS status_vector[20];

  checkBlob();
  if( isc_close_blob(status_vector,
                     &_Handle) )
    throw IBException(__FILE__, __LINE__, status_vector);
  _Handle = NULL;
}

void IBBlobField::create() {
  ISC_STATUS status_vector[20];

  ISC_QUAD* id = new ISC_QUAD();
  isc_db_handle db = ((IBDatabase*)((IBTransaction*)((IBQuery*)_Owner->getQuery())->getTransaction())->getDatabase())->getDBHandle();
  isc_tr_handle tr = ((IBTransaction*)((IBQuery*)_Owner->getQuery())->getTransaction())->getTRHandle();
  if( isc_create_blob2( status_vector,
                        (isc_db_handle*)db,
                        (isc_tr_handle*)tr,
                        &_Handle,
                        id,
                        0,
                        NULL ) )
    throw IBException(__FILE__, __LINE__, status_vector);
  *_var().sqlind = 0;
  _var().sqldata = (char*)id;
}

