/***************************************************************************
                          ibxsqlvar.h  -   SQL
                             -------------------
    begin                : Fri May 12 2000
    copyright            : (C) 2000 by Vadim Morgunov
    email                : vvm@kis.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef IBXSQLVAR_H
#define IBXSQLVAR_H

extern "C" {
#include <ibase.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>
}

#include <iostream>
#include "ibexception.h"
#include "ibasepp_common.h"

using namespace std;

class IBXSQLDA;

/***************************************************************************
        ibase
 ***************************************************************************/

class IBXSQLVAR {
public:
  IBXSQLVAR();
  IBXSQLVAR(IBXSQLDA* sqlda, const int index);
  virtual ~IBXSQLVAR();
public: // methods
  void  setName(const char* newname);
  const char* getName();	
  void setIBXSQLDA(IBXSQLDA* sqlda);
  void setFieldIndex(const int index);
  const int   getType();
  virtual const int   getLength();
  const bool  isNull();
  const bool  isNullable();
  const XSQLVAR* getAsXSQLVAR();
//--------------------------------------------------------------
  virtual operator const char*() ;
  virtual operator short() ;
  virtual operator int()   ;
  virtual operator long()  ;
  virtual operator float() ;
  virtual operator double();
  virtual operator struct tm*();
//--------------------------------------------------------------
  IBXSQLVAR& operator =(const char* value);
  virtual IBXSQLVAR& operator = ( const short value ) ;
  virtual IBXSQLVAR& operator = ( const int value )   ;
  virtual IBXSQLVAR& operator = ( const long value )  ;
  virtual IBXSQLVAR& operator = ( const float value ) ;
  virtual IBXSQLVAR& operator = ( const double value );
  virtual IBXSQLVAR& operator = ( const struct tm* value );
  IBXSQLVAR& operator = (IBXSQLVAR& var);
//--------------------------------------------------------------
public:
  friend ostream& operator<<(ostream& out, IBXSQLVAR& var);
protected:
  IBXSQLDA*  _Owner;
protected: // methods
  virtual char* getAsString() = 0;
  virtual void setAsString( const char* value ) = 0;
  XSQLVAR& _var();
  void checkFieldType(const bool Good);
  void setNull();
private: // fields
  int _FieldIndex;
  char* _StrBuffer;
private:
  void allocateData();
  void clearData();
  void clearStrBuffer();
};

class IBIntField: public IBXSQLVAR {
public:
  ~IBIntField() {};
public:
  operator short();
  IBXSQLVAR& operator = (const short value);
  operator int();
  IBXSQLVAR& operator = (const int value);
  operator long();
  IBXSQLVAR& operator = (const long value);
protected:
  char* getAsString();
  void setAsString( const char* value );
private:
  long getAsInt();
  void setAsInt( const long value );
};

class IBFloatField: public IBXSQLVAR {
public:
  ~IBFloatField() {};
public:
  operator float();
  IBXSQLVAR& operator = (const float value);
  operator double();
  IBXSQLVAR& operator = (const double value);
protected:
  char* getAsString();
  void setAsString( const char* value );
private:
  double getAsFloat();
  void setAsFloat( const double value );
};

class IBStringField: public IBXSQLVAR {
public:
  ~IBStringField() {};
protected:
  char* getAsString();
  void setAsString( const char* value );
  const bool hasWin1251();
  void convertWin2Koi(char** WinStr, int len);
  void convertKoi2Win(char** KoiStr, int len);
};

class IBDateTimeField: public IBXSQLVAR {
public:
  ~IBDateTimeField() {};
public:
  operator struct tm*();
  IBXSQLVAR& operator = (const struct tm* value);
protected:
  char* getAsString();
  void setAsString( const char* value );
private:
  struct tm _Buffer;
private:
  struct tm* getAsDateTime();
  void setAsDateTime( const struct tm* value );
};

class IBBlobField: public IBStringField {
public:
  IBBlobField();
  ~IBBlobField() {};
public:
  const int getLength();
  const int getSubType();
  const ISC_QUAD* getBlobId();
protected:
  void setAsString(const char* value);
  char* getAsString();
private:
  isc_blob_handle _Handle;
  long _Length;
  long _SegmentCount;
  long _SegmentSize;
private:
  void checkBlob();
  void open();
  void close();
  void create();
};

#endif